@@ -4,7 +4,8 @@
 """WebHDFS API clients."""
 
 from .util import AsyncWriter, HdfsError
-from contextlib import contextmanager
+from collections import deque
+from contextlib import closing, contextmanager
 from getpass import getuser
 from itertools import repeat
 from multiprocessing.pool import ThreadPool
@@ -27,13 +28,14 @@ import time
 _logger = lg.getLogger(__name__)
 
 
-def _on_error(response):
+def _to_error(response):
   """Callback when an API response has a non 2XX status code.
 
   :param response: Response.
 
   """
   if response.status_code == 401:
+    _logger.error(response.content)
     raise HdfsError('Authentication failure. Check your credentials.')
   try:
     # Cf. http://hadoop.apache.org/docs/r1.0.4/webhdfs.html#Error+Responses
@@ -41,7 +43,11 @@ def _on_error(response):
   except ValueError:
     # No clear one thing to display, display entire message content
     message = response.content
-  raise HdfsError(message)
+  try:
+    exception = response.json()['RemoteException']['exception']
+  except ValueError:
+    exception = None
+  return HdfsError(message, exception=exception)
 
 
 class _Request(object):
@@ -54,7 +60,7 @@ class _Request(object):
   """
 
   webhdfs_prefix = '/webhdfs/v1'
-  doc_url = 'http://hadoop.apache.org/docs/r1.0.4/webhdfs.html'
+  doc_url = 'https://hadoop.apache.org/docs/stable/hadoop-project-dist/hadoop-hdfs/WebHDFS.html'
 
   def __init__(self, method, **kwargs):
     self.method = method
@@ -75,20 +81,48 @@ class _Request(object):
 
     def api_handler(client, hdfs_path, data=None, strict=True, **params):
       """Wrapper function."""
-      url = '%s%s%s' % (
-        client.url.rstrip('/'),
-        self.webhdfs_prefix,
-        client.resolve(hdfs_path),
-      )
       params['op'] = operation
-      return client._request(
-        method=self.method,
-        url=url,
-        data=data,
-        params=params,
-        strict=strict,
-        **self.kwargs
-      )
+      if client._proxy is not None:
+        params['doas'] = client._proxy
+
+      attempted_hosts = set()
+      while True:
+        with client._lock:
+          while client._urls[0] in attempted_hosts:
+            client._urls.rotate(-1)
+          host = client._urls[0]
+        url = '%s%s%s' % (
+          host.rstrip('/'),
+          self.webhdfs_prefix,
+          quote(client.resolve(hdfs_path), '/= '),
+        )
+
+        err = None
+        try:
+          res = client._request(
+            method=self.method,
+            url=url,
+            data=data,
+            params=params,
+            **self.kwargs
+          )
+        except (rq.exceptions.ReadTimeout, rq.exceptions.ConnectTimeout,
+                rq.exceptions.ConnectionError) as retriable_err:
+          err = retriable_err # Retry.
+        else:
+          if res: # 2XX status code.
+            return res
+          err = _to_error(res)
+          if err.exception not in ('RetriableException', 'StandbyException'):
+            if strict:
+              raise err
+            return res
+
+        attempted_hosts.add(host)
+        if len(attempted_hosts) == len(client._urls):
+          if len(client._urls) > 1:
+            _logger.warning('No reachable host, raising last error.')
+          raise err
 
     api_handler.__name__ = '%s_handler' % (operation.lower(), )
     api_handler.__doc__ = 'Cf. %s#%s' % (self.doc_url, operation)
@@ -123,7 +157,8 @@ class Client(object):
   """Base HDFS web client.
 
   :param url: Hostname or IP address of HDFS namenode, prefixed with protocol,
-    followed by WebHDFS port on namenode.
+    followed by WebHDFS port on namenode.  You may also specify multiple URLs
+    separated by semicolons for High Availability support.
   :param proxy: User to proxy as.
   :param root: Root path, this will be prefixed to all HDFS paths passed to the
     client. If the root is relative, the path will be assumed relative to the
@@ -148,12 +183,12 @@ class Client(object):
   def __init__(self, url, root=None, proxy=None, timeout=None, session=None):
     self.root = root
     self.url = url
+    self.urls = [u for u in url.split(';') if u]
+    self._urls = deque(self.urls) # this is rotated and used internally
     self._session = session or rq.Session()
-    if proxy:
-      if not self._session.params:
-        self._session.params = {}
-      self._session.params['doas'] = proxy
+    self._proxy = proxy
     self._timeout = timeout
+    self._lock = Lock()
     _logger.info('Instantiated %r.', self)
 
   def __repr__(self):
@@ -161,7 +196,7 @@ class Client(object):
 
   # Generic request handler
 
-  def _request(self, method, url, strict=True, **kwargs):
+  def _request(self, method, url, **kwargs):
     """Send request to WebHDFS API.
 
     :param method: HTTP verb.
@@ -171,33 +206,31 @@ class Client(object):
       the instance's defaults.
 
     """
-    response = self._session.request(
+    return self._session.request(
       method=method,
       url=url,
       timeout=self._timeout,
       headers={'content-type': 'application/octet-stream'}, # For HttpFS.
       **kwargs
     )
-    if strict and not response: # Non 2XX status code.
-      return _on_error(response)
-    else:
-      return response
 
   # Raw API endpoints
 
-  _append = _Request('PUT') # doesn't allow for streaming
-  _append_1 = _Request('POST', allow_redirects=False) # cf. `read`
-  _create = _Request('PUT') # doesn't allow for streaming
-  _create_1 = _Request('PUT', allow_redirects=False) # cf. `write`
+  _append = _Request('POST', allow_redirects=False) # cf. `read`
+  _create = _Request('PUT', allow_redirects=False) # cf. `write`
   _delete = _Request('DELETE')
+  _get_acl_status = _Request('GET')
   _get_content_summary = _Request('GET')
   _get_file_checksum = _Request('GET')
   _get_file_status = _Request('GET')
   _get_home_directory = _Request('GET')
+  _get_trash_root = _Request('GET')
   _list_status = _Request('GET')
   _mkdirs = _Request('PUT')
+  _modify_acl_entries = _Request('PUT')
   _open = _Request('GET', stream=True)
   _rename = _Request('PUT')
+  _set_acl = _Request('PUT')
   _set_owner = _Request('PUT')
   _set_permission = _Request('PUT')
   _set_replication = _Request('PUT')
@@ -248,7 +281,7 @@ class Client(object):
     # #LATEST expansion (could cache the pattern, but not worth it)
 
     _logger.debug('Resolved path %r to %r.', hdfs_path, path)
-    return quote(path, '/=')
+    return path
 
   def content(self, hdfs_path, strict=True):
     """Get ContentSummary_ for a file or folder on HDFS.
@@ -280,10 +313,52 @@ class Client(object):
     res = self._get_file_status(hdfs_path, strict=strict)
     return res.json()['FileStatus'] if res else None
 
+  def acl_status(self, hdfs_path, strict=True):
+    """Get AclStatus_ for a file or folder on HDFS.
+
+    :param hdfs_path: Remote path.
+    :param strict: If `False`, return `None` rather than raise an exception if
+      the path doesn't exist.
+
+    .. _AclStatus: ACLS_
+    .. _ACLS: https://hadoop.apache.org/docs/stable2/hadoop-project-dist/hadoop-hdfs/WebHDFS.html#Get_ACL_Status
+
+    """
+    _logger.info('Fetching ACL status for %r.', hdfs_path)
+    res = self._get_acl_status(hdfs_path, strict=strict)
+    return res.json()['AclStatus'] if res else None
+
+  def set_acl(self, hdfs_path, acl_spec, clear=True):
+    """SetAcl_ or ModifyAcl_ for a file or folder on HDFS.
+
+    :param hdfs_path: Path to an existing remote file or directory. An
+      :class:`HdfsError` will be raised if the path doesn't exist.
+    :param acl_spec: String representation of an ACL spec. Must be a valid
+      string with entries for user, group and other. For example:
+      `"user::rwx,user:foo:rw-,group::r--,other::---"`.
+    :param clear: Clear existing ACL entries. If set to false, all existing ACL
+      entries that are not specified in this call are retained without changes,
+      behaving like ModifyAcl_. For example: `"user:foo:rwx"`.
+
+    .. _SetAcl: SETACL_
+    .. SETACL_: https://hadoop.apache.org/docs/stable2/hadoop-project-dist/hadoop-hdfs/WebHDFS.html#Set_ACL
+
+    .. ModifyAcl_: MODACL_
+    .. MODACL_: https://hadoop.apache.org/docs/stable2/hadoop-project-dist/hadoop-hdfs/WebHDFS.html#Modify_ACL_Entries
+
+    """
+    if clear:
+      _logger.info('Setting ACL spec for %r to %r.', hdfs_path, acl_spec)
+      self._set_acl(hdfs_path, aclspec=acl_spec)
+    else:
+      _logger.info('Modifying ACL spec for %r to %r.', hdfs_path, acl_spec)
+      self._modify_acl_entries(hdfs_path, aclspec=acl_spec)
+
   def parts(self, hdfs_path, parts=None, status=False):
     """Returns a dictionary of part-files corresponding to a path.
 
-    :param hdfs_path: Remote path.
+    :param hdfs_path: Remote path. This directory should contain at most one
+      part file per partition (otherwise one will be picked arbitrarily).
     :param parts: List of part-files numbers or total number of part-files to
       select. If a number, that many partitions will be chosen at random. By
       default all part-files are returned. If `parts` is a list and one of the
@@ -373,10 +448,10 @@ class Client(object):
       if permission or blocksize or replication:
         raise ValueError('Cannot change file properties while appending.')
       _logger.info('Appending to %r.', hdfs_path)
-      res = self._append_1(hdfs_path, buffersize=buffersize)
+      res = self._append(hdfs_path, buffersize=buffersize)
     else:
       _logger.info('Writing to %r.', hdfs_path)
-      res = self._create_1(
+      res = self._create(
         hdfs_path,
         overwrite=overwrite,
         permission=permission,
@@ -384,23 +459,25 @@ class Client(object):
         replication=replication,
         buffersize=buffersize,
       )
+    loc = res.headers['location']
 
     def consumer(_data):
       """Thread target."""
-      self._request(
+      res = self._request(
         method='POST' if append else 'PUT',
-        url=res.headers['location'],
+        url=loc,
         data=(c.encode(encoding) for c in _data) if encoding else _data,
-        auth=False,
       )
+      if not res:
+        raise _to_error(res)
 
     if data is None:
       return AsyncWriter(consumer)
     else:
       consumer(data)
 
-  def upload(self, hdfs_path, local_path, overwrite=False, n_threads=1,
-    temp_dir=None, chunk_size=2 ** 16, progress=None, **kwargs):
+  def upload(self, hdfs_path, local_path, n_threads=1, temp_dir=None,
+    chunk_size=2 ** 16, progress=None, cleanup=True, **kwargs):
     """Upload a file or directory to HDFS.
 
     :param hdfs_path: Target HDFS path. If it already exists and is a
@@ -408,7 +485,6 @@ class Client(object):
     :param local_path: Local path to file or folder. If a folder, all the files
       inside of it will be uploaded (note that this implies that folders empty
       of files will not be created remotely).
-    :param overwrite: Overwrite any existing file or directory.
     :param n_threads: Number of threads to use for parallelization. A value of
       `0` (or negative) uses as many threads as there are files.
     :param temp_dir: Directory under which the files will first be uploaded
@@ -419,12 +495,15 @@ class Client(object):
       `chunk_size` bytes. It will be passed two arguments, the path to the
       file being uploaded and the number of bytes transferred so far. On
       completion, it will be called once with `-1` as second argument.
-    :param \*\*kwargs: Keyword arguments forwarded to :meth:`write`.
+    :param cleanup: Delete any uploaded files if an error occurs during the
+      upload.
+    :param \*\*kwargs: Keyword arguments forwarded to :meth:`write`. In
+      particular, set `overwrite` to overwrite any existing file or directory.
 
     On success, this method returns the remote upload path.
 
     """
-    if not chunk_size:
+    if chunk_size <= 0:
       raise ValueError('Upload chunk size must be positive.')
     _logger.info('Uploading %r to %r.', local_path, hdfs_path)
 
@@ -457,20 +536,23 @@ class Client(object):
     try:
       statuses = [status for _, status in self.list(hdfs_path, status=True)]
     except HdfsError as err:
-      if 'not a directory' in str(err):
+      if 'not a directory' in err.message:
         # Remote path is a normal file.
-        if not overwrite:
+        if not kwargs.get('overwrite'):
           raise HdfsError('Remote path %r already exists.', hdfs_path)
-      else:
+      elif 'does not exist' in err.message:
         # Remote path doesn't exist.
         temp_path = hdfs_path
+      else:
+        # An unexpected error occurred.
+        raise err
     else:
       # Remote path is a directory.
       suffixes = set(status['pathSuffix'] for status in statuses)
       local_name = osp.basename(local_path)
       hdfs_path = psp.join(hdfs_path, local_name)
       if local_name in suffixes:
-        if not overwrite:
+        if not kwargs.get('overwrite'):
           raise HdfsError('Remote path %r already exists.', hdfs_path)
       else:
         temp_path = hdfs_path
@@ -480,7 +562,7 @@ class Client(object):
       temp_dir = temp_dir or remote_dpath
       temp_path = psp.join(
         temp_dir,
-        '%s.temp-%s' % (remote_name, int(time.time()))
+        '%s.temp-%s' % (remote_name, _current_micros())
       )
       _logger.debug(
         'Upload destination %r already exists. Using temporary path %r.',
@@ -519,12 +601,15 @@ class Client(object):
       else:
         _map_async(n_threads, _upload, fpath_tuples)
     except Exception as err: # pylint: disable=broad-except
-      _logger.exception('Error while uploading. Attempting cleanup.')
-      try:
-        self.delete(temp_path, recursive=True)
-      except Exception:
-        _logger.error('Unable to cleanup temporary folder.')
-      finally:
+      if cleanup:
+        _logger.exception('Error while uploading. Attempting cleanup.')
+        try:
+          self.delete(temp_path, recursive=True)
+        except Exception:
+          _logger.error('Unable to cleanup temporary folder.')
+        finally:
+          raise err
+      else:
         raise err
     else:
       if temp_path != hdfs_path:
@@ -542,7 +627,7 @@ class Client(object):
 
   @contextmanager
   def read(self, hdfs_path, offset=0, length=None, buffer_size=None,
-    encoding=None, chunk_size=None, delimiter=None, progress=None):
+    encoding=None, chunk_size=0, delimiter=None, progress=None):
     """Read a file from HDFS.
 
     :param hdfs_path: HDFS path.
@@ -575,7 +660,17 @@ class Client(object):
 
     This ensures that connections are always properly closed.
 
+    .. note::
+
+      The raw file-like object returned by this method (when called without an
+      encoding, chunk size, or delimiter) can have a very different performance
+      profile than local files. In particular, line-oriented methods are often
+      slower. The recommended workaround is to specify an encoding when
+      possible or read the entire file before splitting it.
+
     """
+    if chunk_size < 0:
+      raise ValueError('Read chunk size must be non-negative.')
     if progress and not chunk_size:
       raise ValueError('Progress callback requires a positive chunk size.')
     if delimiter:
@@ -672,7 +767,7 @@ class Client(object):
       temp_dir = temp_dir or local_dpath
       temp_path = osp.join(
         temp_dir,
-        '%s.temp-%s' % (local_name, int(time.time()))
+        '%s.temp-%s' % (local_name, _current_micros())
       )
       _logger.debug(
         'Download destination %r already exists. Using temporary path %r.',
@@ -683,7 +778,7 @@ class Client(object):
         raise HdfsError('Parent directory of %r does not exist.', local_path)
       temp_path = local_path
     # Then we figure out which files we need to download and where.
-    remote_paths = list(self.walk(hdfs_path))
+    remote_paths = list(self.walk(hdfs_path, depth=0, status=False))
     if not remote_paths:
       # This is a single file.
       remote_fpaths = [hdfs_path]
@@ -745,22 +840,46 @@ class Client(object):
         )
     return local_path
 
-  def delete(self, hdfs_path, recursive=False):
+  def delete(self, hdfs_path, recursive=False, skip_trash=True):
     """Remove a file or directory from HDFS.
 
     :param hdfs_path: HDFS path.
     :param recursive: Recursively delete files and directories. By default,
       this method will raise an :class:`HdfsError` if trying to delete a
       non-empty directory.
+    :param skip_trash: When false, the deleted path will be moved to an
+      appropriate trash folder rather than deleted. This requires Hadoop 2.9+
+      and trash to be enabled on the cluster.
 
     This function returns `True` if the deletion was successful and `False` if
     no file or directory previously existed at `hdfs_path`.
 
     """
+    verb = 'Deleting' if skip_trash else 'Trashing'
     _logger.info(
-      'Deleting %r%s.', hdfs_path, ' recursively' if recursive else ''
+      '%s %r%s.', verb, hdfs_path, ' recursively' if recursive else ''
     )
-    return self._delete(hdfs_path, recursive=recursive).json()['boolean']
+    if skip_trash:
+      return self._delete(hdfs_path, recursive=recursive).json()['boolean']
+    hdfs_path = self.resolve(hdfs_path)
+    status = self.status(hdfs_path, strict=False)
+    if not status:
+      return False
+    if status['type'] == 'DIRECTORY' and not recursive:
+      raise HdfsError('Non-recursive trashing of directory %r.', hdfs_path)
+    _logger.info('Fetching trash root for %r.', hdfs_path)
+    trash_path = self._get_trash_root(hdfs_path).json()['Path']
+    # The default trash policy (http://mtth.xyz/_9lc9t3hjtz276rx) expects
+    # folders to be under a `"Current"` subfolder. We also add a timestamped
+    # folder as a simple safeguard against path conflicts (note that the above
+    # policy implements a more involved variant).
+    dst_path = psp.join(trash_path, 'Current', _current_micros())
+    self.makedirs(dst_path)
+    # Note that there is a (hopefully small) race condition here: the path might
+    # have been deleted between the status call above and the rename here.
+    self.rename(hdfs_path, dst_path)
+    _logger.info('%r moved to trash at %r.', hdfs_path, dst_path)
+    return True
 
   def rename(self, hdfs_src_path, hdfs_dst_path):
     """Move a file or folder.
@@ -768,7 +887,8 @@ class Client(object):
     :param hdfs_src_path: Source path.
     :param hdfs_dst_path: Destination path. If the path already exists and is
       a directory, the source will be moved into it. If the path exists and is
-      a file, this method will raise an :class:`HdfsError`.
+      a file, or if a parent destination directory is missing, this method will
+      raise an :class:`HdfsError`.
 
     """
     _logger.info('Renaming %r to %r.', hdfs_src_path, hdfs_dst_path)
@@ -896,7 +1016,8 @@ class Client(object):
     else:
       return [s['pathSuffix'] for s in statuses]
 
-  def walk(self, hdfs_path, depth=0, status=False):
+  def walk(self, hdfs_path, depth=0, status=False, ignore_missing=False,
+    allow_dir_changes=False):
     """Depth-first walk of remote filesystem.
 
     :param hdfs_path: Starting path. If the path doesn't exist, an
@@ -904,6 +1025,12 @@ class Client(object):
       generator will be empty.
     :param depth: Maximum depth to explore. `0` for no limit.
     :param status: Also return each file or folder's corresponding FileStatus_.
+    :param ignore_missing: Ignore missing nested folders rather than raise an
+      exception. This can be useful when the tree is modified during a walk.
+    :param allow_dir_changes: Allow changes to the directories' list to affect
+      the walk. For example clearing it by setting `dirs[:] = []` would prevent
+      the walk from entering any nested directories. This option can only be set
+      when `status` is false.
 
     This method returns a generator yielding tuples `(path, dirs, files)`
     where `path` is the absolute path to the current directory, `dirs` is the
@@ -912,20 +1039,38 @@ class Client(object):
 
     """
     _logger.info('Walking %r (depth %r).', hdfs_path, depth)
+    if status and allow_dir_changes:
+      raise ValueError('Cannot set both status and allow_dir_changes')
 
     def _walk(dir_path, dir_status, depth):
       """Recursion helper."""
-      infos = self.list(dir_path, status=True)
+      try:
+        infos = self.list(dir_path, status=True)
+      except HdfsError as err:
+        if ignore_missing and 'does not exist' in err.message:
+          return
+        raise
       dir_infos = [info for info in infos if info[1]['type'] == 'DIRECTORY']
       file_infos = [info for info in infos if info[1]['type'] == 'FILE']
       if status:
         yield ((dir_path, dir_status), dir_infos, file_infos)
       else:
+        dir_names = [dir_name for dir_name, _ in dir_infos]
         yield (
           dir_path,
-          [name for name, _ in dir_infos],
-          [name for name, _ in file_infos],
+          dir_names,
+          [file_name for file_name, _ in file_infos],
         )
+        if allow_dir_changes:
+          infos_by_name = dict(dir_infos)
+          strict = not ignore_missing
+          dir_infos = []
+          for dir_name in dir_names:
+            info = infos_by_name.get(dir_name)
+            if not info:
+              info = self.status(psp.join(dir_path, dir_name), strict=strict)
+            if info:
+              dir_infos.append((dir_name, info))
       if depth != 1:
         for name, s in dir_infos:
           path = psp.join(dir_path, name)
@@ -975,6 +1120,9 @@ class InsecureClient(Client):
     `whoami`).
   :param \*\*kwargs: Keyword arguments passed to the base class' constructor.
 
+  Note that if a session argument is passed in, it will be modified in-place to
+  support authentication.
+
   """
 
   def __init__(self, url, user=None, **kwargs):
@@ -995,6 +1143,9 @@ class TokenClient(Client):
   :param token: Hadoop delegation token.
   :param \*\*kwargs: Keyword arguments passed to the base class' constructor.
 
+  Note that if a session argument is passed in, it will be modified in-place to
+  support authentication.
+
   """
 
   def __init__(self, url, token, **kwargs):
@@ -1008,6 +1159,10 @@ class TokenClient(Client):
 # Helpers
 # -------
 
+def _current_micros():
+  """Returns a string representing the current time in microseconds."""
+  return str(int(time.time() * 1e6))
+
 def _map_async(pool_size, func, args):
   """Async map (threading), handling python 2.6 edge case.
 
@@ -1022,7 +1177,10 @@ def _map_async(pool_size, func, args):
 
   """
   pool = ThreadPool(pool_size)
-  if sys.version_info <= (2, 6):
-    return pool.map(func, args)
-  else:
-    return pool.map_async(func, args).get(1 << 31)
+  with closing(pool):
+    if sys.version_info <= (2, 6):
+      results = pool.map(func, args)
+    else:
+      results = pool.map_async(func, args).get(1 << 22) # 6+ weeks.
+  pool.join()
+  return results
