@@ -25,8 +25,10 @@ class HdfsError(Exception):
 
   """
 
-  def __init__(self, message, *args):
-    super(HdfsError, self).__init__(message % args if args else message)
+  def __init__(self, message, *args, **kwargs):
+    self.message = message % args if args else message
+    super(HdfsError, self).__init__(self.message)
+    self.exception = kwargs.get("exception")
 
 
 class AsyncWriter(object):
@@ -48,6 +50,9 @@ class AsyncWriter(object):
 
   """
 
+  # Expected by pandas to write csv files (https://github.com/mtth/hdfs/pull/130).
+  __iter__ = None
+
   def __init__(self, consumer):
     self._consumer = consumer
     self._queue = None
@@ -70,7 +75,7 @@ class AsyncWriter(object):
         _logger.debug('Starting consumer.')
         self._consumer(data)
       except Exception as err: # pylint: disable=broad-except
-        _logger.debug('Exception in child.')
+        _logger.exception('Exception in child.')
         self._err = err
       finally:
         _logger.debug('Finished consumer.')
@@ -105,6 +110,18 @@ class AsyncWriter(object):
     """Pass-through implementation."""
     pass
 
+  def seekable(self):
+    """Implement file-like method expected by certain libraries.
+
+    `fastavro` relies on it in python 3.
+
+    """
+    return False
+
+  def tell(self):
+    """No-op implementation."""
+    return 0
+
   def write(self, chunk):
     """Stream data to the underlying consumer.
 
@@ -112,7 +129,11 @@ class AsyncWriter(object):
       consumer reads them.
 
     """
-    self._queue.put(chunk)
+    if chunk:
+      # We skip empty chunks, otherwise they cause request to terminate the
+      # response stream. Note that these chunks can be produced by valid
+      # upstream encoders (e.g. bzip2).
+      self._queue.put(chunk)
 
 
 @contextmanager
